{-# LANGUAGE LambdaCase, TypeSynonymInstances, FlexibleInstances, 
             IncoherentInstances, TypeFamilies, CPP #-}

-- August 4, 2021

-- The page numbers refer to https://fldit-www.cs.tu-dortmund.de/~peter/
-- DialgSlides.pdf

module Coalg where

import Data.Monoid (Monoid(mempty,mappend))
import Data.List (delete,(\\))
import Control.Applicative (Applicative(pure,(<*>)),liftA2,
			    Alternative(empty,(<|>)))
import Data.Array(Ix,Array,array,bounds,range,(!))
import Data.Maybe (fromJust)
import Data.Either (either)
import Control.Monad 
import Painter (indices,meet,Tree(V,F),root,subtrees,height,mapTree,mkTrees,
		update,MonadTrans(type M,lift,liftT),MaybeT(MaybeT,runMT))

#if __GLASGOW_HASKELL__ < 800
instance Monoid state => Monad ((,) state) where               
         return a = (mempty,a)                          
         (st,a) >>= f = (st `mappend` st',b) where (st',b) = f a 
#endif

#if __GLASGOW_HASKELL__ >= 804
instance Semigroup Int where (<>) = mappend
#endif

#if __GLASGOW_HASKELL__ >= 806
import Control.Monad.Fail hiding (fail)		
import qualified Control.Monad.Fail (fail)	
instance MonadFail m => MonadFail (StateT state m) where fail = fail
#endif

type Node = [Int]

(***) :: (a -> c) -> (b -> d) -> (a,b) -> (c,d)		
(f***g) (a,b) = (f a,g b)

(+++) :: (a -> c) -> (b -> d) -> Either a b -> Either c d 
f+++g = either (Left . f) $ Right . g

(?) :: (a -> Bool) -> a -> Either a a
p?a = if p a then Left a else Right a

ite :: (a -> Bool) -> (a -> b) -> (a -> b) -> a -> b
ite p f g = either f g . (p?)          -- ite p f g a = if p a then f a else g a             

class Contravariant f where contramap :: (a -> b) -> f b -> f a

newtype Coreader state a = CR {runCR :: a -> state}

instance Contravariant (Coreader state) where contramap f (CR h) = CR $ h . f

(<>) :: (a -> b) -> (a -> c) -> a -> (b,c)
(<>) = liftA2 (,)

foo = ((+1) <> (+3)) 5 -- > (6,8)

(+:),(*:) :: (Num a,Applicative f) => f a -> f a -> f a
(+:) = liftA2 (+)
(*:) = liftA2 (*)
         
-- identity monad

newtype Id a = Id {run :: a}

instance Functor Id where fmap = liftM

instance Applicative Id where pure  = return
         		      (<*>) = liftM2 id

instance Monad Id where return = Id			
			Id a >>= f = f a						
         
{- reader monads

instance Functor ((->) state) where
 	 fmap f h = f . h

instance Monad ((->) state) where			
         return = const
         (h >>= f) st = f (h st) st

-- writer monads (state is a monoid)
         
instance Functor ((,) state) where			
 	 fmap f (st,a) = (st,f a)
 	 
class Monoid a where mempty :: a
		     mappend :: a -> a -> a
		     
mconcat :: Monoid a => [a] -> a
mconcat = foldr mappend mempty -}

-- term monad

data Term a b = FT a [Term a b] | VT b deriving Show

instance Functor (Term a) where fmap f (FT a ts) = FT a $ map (fmap f) ts 
                  	        fmap f (VT b)    = VT $ f b 

instance Applicative (Term a) where pure  = return
         			    (<*>) = liftM2 id

instance Monad (Term a) where			
               return = VT
               t >>= sub = case t of FT a ts -> FT a $ map (>>= sub) ts 
                                     VT b -> sub b

data Vars = X | Y | Z | X1 | Y1 | Z1 deriving (Eq,Show)
 
t1 :: Term String Vars
t1 = FT"+"[FT"*"[FT"5"[],VT X],VT Y, FT"11"[]]

sub1 :: Vars -> Term String Vars
sub1 = \case X -> FT"/"[FT"-"[FT"6"[]],FT"9"[],VT Z]   
             Y -> FT"-"[FT"7"[],FT"*"[FT"8"[],FT"0"[]]]  
             x -> VT x

t2 = t1 >>= sub1 -- > FT"+"[FT"*"[FT"5"[],FT"/"[FT"-"[FT"6"[]],FT"9"[],VT Z]],
                 --         FT"-"[FT"7"[],FT"*"[FT"8"[],FT"0"[]]],
                 --         FT"11"[]]         

notIn :: Eq b => b -> Term a b -> Bool
b `notIn` VT b'   = b /= b'
b `notIn` FT _ ts = all (b `notIn`) ts
 
unify :: (Eq a,Eq b) => Term a b -> Term a b -> Maybe (b -> Term a b)
unify (VT b) (VT c)        = Just $ if b == c then VT else update VT b $ VT c
unify (VT b) t             = do guard $ b `notIn` t
			        Just $ update VT b t
unify t (VT b)             = unify (VT b) t
unify (FT a ts) (FT a' us) = do guard $ a == a' && length ts == length us
                                unifyList ts us

unifyList :: (Eq a,Eq b) => [Term a b] -> [Term a b] -> Maybe (b -> Term a b)
unifyList [] []         = Just VT                  
unifyList (t:ts) (u:us) = do sub <- unify t u
                             let msub = map (>>= sub)
                             sub' <- unifyList (msub ts) $ msub us
                             Just $ sub >=> sub'

sub2 = fromJust $ unify (FT"+"[FT"-"[VT X,VT Y],VT Z]) 
			(FT"+"[VT X1,FT"*"[VT Y1,VT Z1]])  
                          		-- > sub2 Z  = FT"*"[VT Y1,VT Y2]
                          		--   sub2 X1 = FT"-"[VT X,VT Y]
                          		--   sub2 x  = VT x}
 
-- state monad
         
newtype State state a = State {runS :: state -> (a,state)}

instance Functor (State state) where fmap = liftM

instance Applicative (State state) where pure  = return
         			         (<*>) = liftM2 id

instance Monad (State state) where			
               return a = State $ \st -> (a,st)
               State h >>= f = State $ (\(a,st) -> runS (f a) st) . h
                                     
-- state transformer monad

newtype StateT state m a = StateT {runST :: state -> m (a,state)}

instance Monad m => Functor (StateT state m) where fmap = liftM 
      -- f (StateT h) = StateT $ (\(a,st) -> return (f a,st)) <=< h
	 
instance Monad m => Applicative (StateT state m) where pure = return
						       (<*>) = liftM2 id

instance Monad m => Monad (StateT state m) where  	
         return a = StateT $ \st -> return (a,st)	        
         StateT h >>= f = StateT $ (\(a,st) -> runST (f a) st) <=< h
         fail = StateT . const . fail		   

instance MonadPlus m => Alternative (StateT state m) where
 	 empty = mzero
 	 (<|>) = mplus
	       
instance MonadPlus m => MonadPlus (StateT state m) where
         mzero = StateT $ const mzero
         StateT g `mplus` StateT h = StateT $ liftM2 mplus g h
 	 					
instance MonadTrans (StateT state) where
		    type M (StateT state) = State state
                    lift m = StateT $ \st -> do a <- m; return (a,st)
                    		         -- :: m a -> StateT state m a
                    liftT (State f) = StateT $ return . f  
                    		         -- :: State state a -> StateT state m a

-- basic compilers

type Compiler = StateT String Maybe

parse :: Compiler a -> String -> Maybe a
parse p str = do (a,_) <- runST p str; Just a

sat :: (Char -> Bool) -> Compiler Char
sat f = StateT $ \str -> do c:str <- return str
                            if f c then return (c,str) else mzero

char :: Char -> Compiler Char
char chr = sat (== chr)

digit :: Compiler Char
digit  = msum $ map char ['0'..'9']

some, many :: MonadPlus m => m a -> m [a]
some p = do a <- p; as <- many p; return $ a:as
many p = msum [some p, return []]

delim :: Compiler Char
delim  = msum $ map char " \n\t"

token :: Compiler a -> Compiler a
token comp = do many delim; a <- comp; many delim; return a

tchar = token . char

nat :: Compiler Int
nat = fmap read $ some digit

-- comonad class

infixl 1 <<=

class Functor cm => Comonad cm where extract :: cm a -> a
     			             (<<=) :: (cm a -> b) -> (cm a -> cm b)

(=<=) :: Comonad cm => (cm b -> c) -> (cm a -> b) -> (cm a -> c)
g =<= f = g . (f <<=)   

duplicate :: Comonad cm => cm a -> cm (cm a)
duplicate = (id <<=)

-- identity comonad

instance Comonad Id where				
         extract = run
         f <<= cm = Id $ f cm

-- reader comonads (state is a monoid)

instance Monoid state => Comonad ((->) state) where 	
         		 extract g = g mempty				
         		 (f <<= g) st = f $ g . mappend st 

-- writer comonads

instance Comonad ((,) state) where extract (_,a) = a
        			   f <<= p@(st,_) = (st,f p)
         
-- costate comonads 

infixr 4 :#

data Costate state a = (:#) {out :: state -> a, final :: state}
							
instance Functor (Costate state) where fmap f (h:#st) = (f . h):#st
     
instance Comonad (Costate state) where extract (h:#st) = h st
         			       f <<= (h:#st) = (f . (:#) h):#st 

type InflistNode = Costate Int

type InftreeNode = Costate Node


-- NATURAL NUMBERS

data NatSig val = NAS {zero :: val, suc :: val -> val}  -- signature

data Nat = Zero | Succ Nat				-- terms
  
one,infN :: Nat
one  = Succ Zero
infN = Succ infN

natT :: NatSig Nat					-- term algebra
natT = NAS Zero Succ
             
foldNat :: NatSig val -> Nat -> val			  
foldNat alg Zero     = zero alg
foldNat alg (Succ n) = suc alg $ foldNat alg n
               
foldInt :: NatSig val -> Int -> val			  
foldInt alg 0         = zero alg
foldInt alg n | n > 0 = suc alg $ foldInt alg $ n-1
  
toInt :: Nat -> Int
toInt Zero     = 0
toInt (Succ n) = toInt n+1

instance Show Nat where show = show . toInt
                            
-- Inductive definitions on Nat or Int

caseN :: x -> (Nat -> x) -> Nat -> x
caseN x _ Zero     = x
caseN _ f (Succ n) = f n

add,mul :: Nat -> Nat -> Nat
add Zero n     = n
add (Succ m) n = Succ $ add m n
mul Zero n     = Zero
mul (Succ m) n = add (mul m n) n
				
predN :: Nat -> Nat					-- predecessor
predN = fst . d where d :: Nat -> (Nat,Nat)		-- see p. 414
        	      d Zero     = (Zero,Zero)
        	      d (Succ n) = suc $ d n 
        	      		   where suc :: (Nat,Nat) -> (Nat,Nat)
        	      		         suc (_,n) = (n,Succ n)
        	      
predI =  predN . toNat 

factN :: Nat -> Nat 					-- factorial
factN = fst . d where d :: Nat -> (Nat,Nat)		-- see p. 417
        	      d Zero     = (one,Zero)
        	      d (Succ n) = suc $ d n 
        	      		   where suc :: (Nat,Nat) -> (Nat,Nat)
        	      		         suc (m,n) = (mul m $ Succ n,Succ n)
toNat :: Int -> Nat
toNat 0         = Zero
toNat n | n > 0 = Succ $ toNat $ n-1

factI = factN . toNat 

fact,factF :: Int -> Int

fact 0         = 1
fact n | n > 0 = fact (n-1)*n

factF = fst . foldInt factAlg

factAlg :: NatSig (Int,Int)		          
factAlg = NAS (1,0) $ \(m,n) -> (m*(n+1),n+1)

subS,subF :: Int -> Int -> Int				-- subtraction

subS 0 n = 0
subS m 0 = m
subS m n = subS (m-1) $ n-1

subF = fst . foldInt subAlg

subAlg :: NatSig (Int -> Int,Int)
subAlg = NAS {zero = (const 0,0), 
              suc  = \(f,m) -> (aux f m,m+1)} where aux f m 0 = m+1		
       	      				            aux f m n = f $ n-1

subN :: Nat -> Nat -> Nat				-- see p. 418
subN = fst . d where 
       d :: Nat -> (Nat -> Nat,Nat)
       d Zero     = (const Zero,Zero)
       d (Succ n) = suc $ d n where suc :: (Nat -> Nat,Nat) -> (Nat -> Nat,Nat)
      		     	            suc (f,m) = (aux f m,Succ m) 
       aux f m Zero     = Succ m		
       aux f m (Succ n) = f n
        	      
subI m = subN (toNat m) . toNat
              			   
ack,ackN :: Nat -> (Nat -> Nat)

ack Zero n            = Succ n
ack (Succ m) Zero     = ack m one
ack (Succ m) (Succ n) = ack m $ ack (Succ m) n

ackN Zero     = Succ					-- see p. 419 f.
ackN (Succ n) = suc $ ackN n where suc :: (Nat -> Nat) -> (Nat -> Nat)
	                           suc f Zero     = f one		
	                           suc f (Succ n) = f $ suc f n

ackI m = toInt . ackN (toNat m) . toNat 

-- Natural numbers with infinity

data ConatSig val = CNS {pre :: val -> Either () val}	-- signature

newtype Conat = CN {predC :: Either () Conat} 		-- coterms

zeroC,infC :: Conat
zeroC = CN $ Left ()				
infC  = CN $ Right infC

succC :: Conat -> Conat
succC = CN . Right 					-- see p. 436
 
conat :: ConatSig Conat			                -- coterm algebra
conat = CNS predC

unfoldConat :: ConatSig val -> val -> Conat	
unfoldConat alg val = either (const zeroC) (succC . unfoldConat alg) 
                             $ pre alg val

natC :: ConatSig Nat				        -- final Conat-algebra 
natC = CNS $ caseN (Left ()) Right			-- with term carrier

unfoldNat :: ConatSig val -> val -> Nat  		
unfoldNat alg val = either (const Zero) (Succ . unfoldNat alg) $ pre alg val

succI = succC . toConat 
                          
toConat :: Int -> Conat
toConat 0         = zeroC
toConat n | n > 0 = succC $ toConat $ n-1
                   
toIntC :: Conat -> Int
toIntC (CN (Left _))  = 0
toIntC (CN (Right n)) = toIntC n+1

instance Show Conat where show = show . toIntC
                            
-- Coinductive definitions to Conat or Nat

lengthC :: Colist x -> Conat				-- see p. 437
lengthC = CN . (id+++lengthC) . pre
          where pre s = either (Right . snd) (const $ Left ()) $ splitC s

lengthI = lengthC . toColist

addCC :: Conat -> Conat -> Conat
addCC m n = case predC m of Left _ -> case predC n of 
					   Left _ -> zeroC
					   Right n' -> succC $ addCC m n'
                            Right m' -> succC $ addCC m' n

addCCI m = addCC (toConat m) . toConat
			 
addC :: (Conat,Conat) -> Conat				-- see p. 437 f.
addC = CN . (id+++addC) . pre
       where pre (m,n) = either (const $ either (const $ Left ()) 
       					        (\n' -> Right (m,n')) $ predC n)
                                (\m' -> Right (m',n)) $ predC m

addCI m n = addC (toConat m,toConat n)


-- SEQUENCES

data ListSig x val = LIS {nil :: val, 			-- signature
			  cons :: x -> val -> val} 

listT :: ListSig x [x]				       	-- term algebra
listT = LIS [] (:)

foldList :: ListSig x val -> [x] -> val			 
foldList alg = \case x:s -> cons alg x $ foldList alg s
		     _ -> nil alg

foldList' alg = foldr (cons alg) $ nil alg

list1 = take 1000 blink where blink = 0:1:blink		 

list2 = take 1000 $ iterate (+2) 22 
list3 = take 1000 $ iterate f 0 where f x = (x+1) `mod` 2	       

intAlg :: ListSig Int Int			      	-- List-algebra 
intAlg = LIS 0 (+)					-- with carrier Int

list4 = foldList intAlg [1..8]
list5 = foldr (+) 0 [1..8]
list6 = foldList' intAlg [1..8]
                            
-- Inductive definitions on [x] 

flipZipWith :: [x] -> (x -> y -> z) -> [y] -> [z] 	-- see p. 423 ff.
flipZipWith (x:s) = cons x $ flipZipWith s
	            where cons :: x -> ((x -> y -> z) -> [y] -> [z])
	                            -> ((x -> y -> z) -> [y] -> [z])
	                  cons x h' = \g -> aux x g h'
	                  aux x g h' []     = []
	                  aux x g h' (y:s') = g x y:h' g s' 
flipZipWith _     = \g -> \s' -> []

sorted,sortedL :: Ord x => [x] -> Bool

sorted (x:y:s) = x <= y && sorted (y:s)
sorted _       = True

sortedL = fst . d' where 
          d' :: Ord x => [x] -> (Bool,[x])
          d' (x:s) = cons x $ d' s where		-- see p. 428
		     cons :: Ord x => x -> (Bool,[x]) -> (Bool,[x])
		     cons x (_,s) = (d s x,x:s)
	  d' _     = (True,[])
	  d :: Ord x => [x] -> x -> Bool
	  d (x:s)  = cons x $ d s where 
	  	     cons :: Ord x => x -> (x -> Bool) -> x -> Bool
	  	     cons x p y = x <= y && p y
          d _      = const True
		         
subtree :: Node -> (Bintree x -> Bintree x)		-- see p. 429
subtree (b:s) = cons b $ subtree s where
                cons :: Int -> (Bintree x -> Bintree x) 
                	    -> (Bintree x -> Bintree x)
		cons b f (Bjoin _ t u) = if b == 0 then f t else f u
	        cons b f _ = Empty
subtree _ = id

subtree1 = subtree [0,1,0] baltree1			-- > 8(7)
		         
subtreeC :: Node -> (Cobintree x -> Cobintree x)	-- see p. 430
subtreeC (b:s) = cons b $ subtreeC s where
                 cons :: Int -> (Cobintree x -> Cobintree x) 
                             -> (Cobintree x -> Cobintree x)
		 cons b f t = either (\(_,t,u) -> if b == 0 then f t else f u) 
			             (const emptyC) $ splitBC t
subtreeC _ = id

subtree2 = subtreeC [0,1,0] $ toCobintree baltree1	-- > 8(7)

instance Num a => Num [a] where
         (+) = zipWith (+)  --  = (+:)
	 s*s' = x*head s':tail s*s'+(x:0)*tail s' 	-- convolution product
	        where x = head s
	 negate s = negate (head s):negate (tail s)
	 abs = id
	 signum _ = 1
	 fromInteger n = fromInteger n:0
	 
instance Fractional a => Fractional [a] where
         recip s = x:negate ((x:0)*tail s)*recip s where x = recip $ head s
	 fromRational r = fromRational r:0

sucs  = 1:ones+sucs where ones = 1:ones

facts = 1:zipWith (*) facts sucs

fibs  = 0:fibs+(1:fibs)
fibs0 = 1:1:fibs0+tail fibs0
fibs1 = 0:tailfibs where tailfibs = 1:fibs1+tailfibs
fibs2 = 0:y*fibs1+z where y = 1:1; z = 1:0
fibs3 = x*recip (z-x-x*x) where x = 0:1; z = 1:0

{- 
instance Functor [ ] where fmap = map
      
instance Monad [ ] where return a = [a]
       		         (>>=) = flip concatMap
      
instance MonadPlus [ ] where mzero = []
         		     mplus = (++)
-}

instance Comonad [ ] where extract = head
         		   f <<= [] = []
         		   f <<= s  = f s:(f <<= tail s)

list7 = id <<= [1..5]		-- > [[1,2,3,4,5],[2,3,4,5],[3,4,5],[4,5],[5]]
list8 = length <<= [1..8]	-- > [8,7,6,5,4,3,2,1]
list9 = sum <<= [1..8]		-- > [36,35,33,30,26,21,15,8]

-- Streams
         
data StreamSig x val = STS {head_ :: val -> x, 
			    tail_ :: val -> val}      -- signature
infixr 5 :<

data Stream x = (:<) {hd :: x, tl :: Stream x} 	      -- coterms

streamC :: StreamSig x (Stream x)		      -- coterm algebra
streamC = STS hd tl

unfoldStream :: StreamSig x val -> val -> Stream x 	 
unfoldStream alg val = head_ alg val :< unfoldStream alg (tail_ alg val)

toStream :: [x] -> Stream x
toStream (x:s) = x :< toStream s

streamFun :: StreamSig x (Int -> x)		      -- final Stream-algebra
streamFun = STS ($0) (\s n -> s $ n+1)	              -- with functional carrier

unfoldFun :: StreamSig x val -> val -> (Int -> x)        
unfoldFun alg val = \case 0 -> head_ alg val 
			  n -> unfoldFun alg (tail_ alg val) $ n-1

stream1 = map (unfoldFun streamC blink$) [0,1,500,501,1200,1205]
	 			  
(!!!) :: Stream x -> Int -> x 
(x :< _)!!!0 = x
(_ :< s)!!!n = s!!!(n-1)

prefix :: Int -> Stream x -> [x]
prefix 0 s      = []
prefix n (x:<s) = x:prefix (n-1) s

instance Show x => Show (Stream x) where show = show . prefix 100

instance Functor Stream where 
	 fmap f (x :< s) = f x :< fmap f s

instance Applicative Stream where 
	 pure x = s where s = x :< s
	 (f :< fs) <*> (x :< s) = f x :< (fs <*> s)   -- zipWith ($)

instance Num x => Num (Stream x) where
         (+) = zipW (+)  --  = (+:)
	 (*) = zipW (*)  --  = (*:) 
	 negate s = negate (hd s) :< negate (tl s)
	 abs = id
	 signum _ = 1
	 fromInteger n = fromInteger n:<0
	 
instance Fractional x => Fractional (Stream x) where
         recip s = x :< negate ((x:<0)*tl s)*recip s where x = recip $ hd s
	 fromRational r = fromRational r :< 0		          

nats n = n:<nats(n+1)

consI :: a -> (Int -> a) -> Int -> a
consI a f = \case 0 -> a; n -> f $ n-1

blinkI = consI 0 $ consI 1 blinkI
natsI n = consI n $ natsI $ n+1

sumC :: Num x => Stream (Stream x) -> Stream x
sumC ss = hd ss+sumC (tl ss)

-- Coinductive definitions to Stream x
			   
ones'  = 1:<ones'
twos   = 2:<twos
sucs'  = 1:<ones'+sucs'
nats'  = 0:<nats'+ones'
nats1  = (0:<nats1)+ones'

evennats = 0:<oddnats+ones' 
oddnats  = 1:<oddnats+twos

repC n = n:<repC n

evens s = hd s:<evens (tl $ tl s)

facts1 = 1:<facts1*sucs'
facts2 = 1:<(nats'+ones')*facts2

fibs4 = 0:<fibs4+(1:<fibs4)
fibs5 = 0:<tailfibs where tailfibs = 1:<fibs5+tailfibs
fibs6 = 1:<fibs6+(1:<fibs6)
	 			  
mapS :: (a -> b) -> Stream a -> Stream b
mapS f s = f (hd s) :< mapS f (tl s)
	 			  
zipS :: Stream a -> Stream a -> Stream a		-- see p. 440
zipS s s' = hd s :< zipS s' (tl s)

zipL :: [a] -> [a] -> Stream a
zipL s = zipS (toStream s) . toStream

-- zipL [1,3..] [2,4..]	-> [1..100]

blink,blink' :: Stream Int				     
blink  = 0 :< blink'
blink' = 1 :< blink

blinkC :: () -> Stream Int				-- see p. 441
blinkC = c . Left where c :: Either () () -> Stream Int
                        c sum = either (const 0) (const 1) sum :<
                                c (either Right Left sum)
                                
-- **** appzeros see p. 442                               

exchange :: Stream x -> Stream x			-- see p. 425 f.
exchange = c . Left where c :: Either (Stream x) (Stream x) -> Stream x
		          c sum = either (hd . tl) hd sum :<
		                  c (either Right (Left . tl . tl) sum)
		         
exchangeL :: [x] -> Stream x
exchangeL = exchange . toStream

-- exchangeL [1..] -> [2,1,4,3,6,5,8,7,10,9,12,11,14,13,16,15,18,17,...,100,99]

nats_squares :: Int -> Stream Int				-- see p. 426 f.
nats_squares = c . Left where c :: Either Int Int -> Stream Int
		              c sum = either id (\n -> n*n) sum :<
		                      c (either Right (Left . (+1)) sum)
		                      
-- nats_squares 5 -> [5,25,6,36,7,49,8,64,9,81,10,100,11,121,12,144,13,169,...]
	 			  
zipW :: (a -> b -> c) -> Stream a -> Stream b -> Stream c       -- see p. 427 f.
zipW f s s' = f (hd s) (hd s') :< zipW f (tl s) (tl s')
		             
foldprefixes :: (y -> x -> y) -> y -> Stream x -> Stream y	-- see p. 428
foldprefixes f y s = f y (hd s) :< foldprefixes f (f y $ hd s) (tl s)

cycleNats :: Int -> Stream Int
cycleNats n  = c n n where c m n = m :< case m of 0 -> c n n
			                          _ -> c (m-1) n

cycleNatsN :: Nat -> Stream Nat					-- see p. 428 f.
cycleNatsN n = c (n,n) where c :: (Nat,Nat) -> Stream Nat
			     c (m,n) = m :< c (caseN (n,n) (\m -> (m,n)) m)
			                          
cycleNatsI :: Int -> Stream Int	
cycleNatsI = mapS toInt . cycleNatsN . toNat
			                          			       
-- cycleNats n n = concat $ repeat $ reverse [0..n]
-- cycleNats{I} 4 -> [4,3,2,1,0,4,3,2,1,0,4,3,2,1,0,4,3,2,1,0,4,3,2,1,0,...]

merge :: Ord x => (Stream x,Stream x) -> Stream x		-- see p. 447 f.
merge (s,s') = min hds hds' :< merge (if hds < hds' then (tls,s') 
			              else if hds == hds' then (tls,tls') 
			                                  else (s,tls'))
	       where hds :< tls = s; hds' :< tls' = s' 
	       
mergeL :: Ord x => [x] -> [x] -> Stream x
mergeL s s' = merge (toStream s,toStream s')

-- mergeL [1,3..] [2,4..] -> [1..100]

insert :: Ord x => (x,Stream x) -> Stream x			-- see p. 432 f.
insert = c . Left where 
         c :: Ord x => Either (x,Stream x) (Stream x) -> Stream x
         c sum = either (\(x,s) -> min x $ hd s) hd sum :<
                 c (either (\(x,s) -> if x <= hd s then Right s
		                	           else Left (x,tl s))
		           (Right . tl) sum)
		           
insertL :: Ord x => x -> [x] -> Stream x
insertL x s = insert (x,toStream s)

-- insertL 12 [1,3..] -> [1,3,5,7,9,11,12,13,15,17,19,21,23,25,27,29,31,33,...]

instance Monoid Int where mempty = 0
			  mappend = (+)

listF :: (Int -> x) -> [x]
listF = flip map [0..]

repeatF :: x -> Int -> x
repeatF = return

stream2 = take 11 $ listF $ repeatF 45

headF :: (Int -> x) -> x
headF = extract				

dropF :: Int -> (Int -> x) -> Int -> x
dropF i s = ($i) <<= s				      -- ($i) :: (Int -> x) -> x

tailF :: (Int -> x) -> Int -> x
tailF = dropF 1				

fby :: (Int -> x) -> (Int -> x) -> Int -> x
fby s _ 0 = s $ 0
fby _ s i = s $ i-1

consF :: x -> (Int -> x) -> Int -> x
consF x xs = repeatF x `fby` xs

zipWith1,zipWith2 :: (a -> b -> c) -> (Int -> a) -> (Int -> b) -> Int -> c
zipWith1 f as bs = do a <- as; b <- bs; return $ f a b
zipWith2 f as bs = consF (f (headF as) $ headF bs) $
		         zipWith2 f (tailF as) $ tailF bs

ones,pnats :: Int -> Int
ones = repeatF 1
pnats = ones `fby` zipWith2 (+) ones pnats

stream3 = take 11 $ listF $ dropF 15 pnats
stream4 = take 11 $ listF $ fmap (*3) pnats     
stream5 = take 11 $ listF $ zipWith1 (+) pnats pnats
stream6 = take 11 $ listF $ zipWith2 (+) pnats pnats

prefixSum1,neighbSum1 :: (Int -> Int) -> Int -> Int
prefixSum1 s 0 = s 0
prefixSum1 s i = prefixSum1 s (i-1)+s i
neighbSum1 s 0 = s 0+s 1
neighbSum1 s i = s (i-1)+s i+s (i+1)

stream7 = take 8 $ listF $ prefixSum1 pnats	-- > [1,3,6,10,15,21,28,36]
stream8 = take 8 $ listF $ neighbSum1 pnats  	-- > [3,6,9,12,15,18,21,24]

-- Lists with a position

infixr 4 :@

data ListPos a = (:@) {list :: [a], pos :: Int}	

instance Functor ListPos where fmap f (s:@i) = map f s:@i
      
instance Comonad ListPos where extract (s:@i) = s!!i
			       f <<= s:@i = map (f . (s:@)) (indices s):@i 
         
prefixSum,suffixSum,neighbSum :: ListPos Int -> Int
prefixSum (s:@i) = sum $ take (i+1) s
suffixSum (s:@i) = sum $ drop i s 
neighbSum (s:@0) = s!!0+s!!1
neighbSum (s:@i) | i < length s-1 = s!!(i-1)+s!!i+s!!(i+1)
	         | True  	  = s!!(i-1)+s!!i

list11 = list $ prefixSum <<= [1..8]:@0         -- >  [1,3,6,10,15,21,28,36]
list12 = list $ suffixSum <<= [1..8]:@0         -- >  [36,35,33,30,26,21,15,8]
list13 = list $ neighbSum <<= [1..8]:@0         -- >  [3,6,9,12,15,18,21,15]

-- Matrices with a position

type Pos = (Int,Int)

type Matrix = Array Pos

dim :: Matrix r -> Int
dim = fst . snd . bounds

mkArray :: Ix a => (a,a) -> (a -> b) -> Array a b 
mkArray bds f = array bds [(x,f x) | x <- range bds] 

mkMat :: Int -> (Pos -> r) -> Matrix r
mkMat d = mkArray ((1,1),(d,d))
		  
instance Show (Matrix Int) where 
         show mat = "    " ++ concatMap f [1..d] ++ 
                    "\n    " ++ replicate (3*d) '-' ++ 
                    '\n':concatMap showRow [1..d]
              where d = dim mat
                    f i = if i < 10 then "  "++show i else ' ':show i
                    showRow i = f i++'|':concatMap (showEntry i) [1..d]++"\n"
                    showEntry i j = if e < 10 then "  "++show e else ' ':show e
                                    where e = mat!(i,j)

mat1,mat2 :: Matrix Int 
mat1 = mkMat 10 $ \(i,j) -> if i == j then 11 else 0
mat2 = mkMat 10 $ \(i,j) -> i+j

{-    1  2  3  4  5  6  7  8  9 10
    ------------------------------
  1|  2  3  4  5  6  7  8  9 10 11
  2|  3  4  5  6  7  8  9 10 11 12
  3|  4  5  6  7  8  9 10 11 12 13
  4|  5  6  7  8  9 10 11 12 13 14
  5|  6  7  8  9 10 11 12 13 14 15
  6|  7  8  9 10 11 12 13 14 15 16
  7|  8  9 10 11 12 13 14 15 16 17
  8|  9 10 11 12 13 14 15 16 17 18
  9| 10 11 12 13 14 15 16 17 18 19
 10| 11 12 13 14 15 16 17 18 19 20 -}

infixr 4 :%

data MatPos a = (:%) {matrix :: Matrix a, pos2 :: Pos}	

instance Functor MatPos where fmap f (mat:%pos) = fmap f mat:%pos
    
instance Comonad MatPos where 
         extract (mat:%pos) = mat!pos
         f <<= mat:%pos = mkMat (dim mat) (f . (mat:%)):%pos
         		      		       
neighbSum2 :: MatPos Int -> Int
neighbSum2 (mat:%(i,j)) = sum $ map (mat!) 
			      $ meet [(i-1,j),(i,j-1),(i,j),(i,j+1),(i+1,j)]
			      $ range $ bounds mat

mat3 = matrix $ neighbSum2 <<= mat2:%(1,1)

{-    1  2  3  4  5  6  7  8  9 10
    ------------------------------
  1|  8 13 17 21 25 29 33 37 41 33
  2| 13 20 25 30 35 40 45 50 55 47
  3| 17 25 30 35 40 45 50 55 60 51
  4| 21 30 35 40 45 50 55 60 65 55
  5| 25 35 40 45 50 55 60 65 70 59
  6| 29 40 45 50 55 60 65 70 75 63
  7| 33 45 50 55 60 65 70 75 80 67
  8| 37 50 55 60 65 70 75 80 85 71
  9| 41 55 60 65 70 75 80 85 90 75
 10| 33 47 51 55 59 63 67 71 75 58 -} 
  				  
-- Colists

data ColistSig x val = CLS {split :: val -> Either (x,val) ()} -- signature

newtype Colist x = CL {splitC :: Either (x,Colist x) ()}       -- coterms

colistC :: ColistSig x (Colist x)			       -- coterm algebra
colistC = CLS splitC

consC :: x -> Colist x -> Colist x
consC x s = CL $ Left (x,s)

nilC :: Colist x
nilC = CL $ Right ()

unfoldColist :: ColistSig x val -> val -> Colist x	      
unfoldColist alg val = either (\(x,val) -> consC x $ unfoldColist alg val)
			      (const nilC) $ split alg val 

toColist :: [x] -> Colist x
toColist (x:s) = consC x $ toColist s
toColist _     = nilC

prefixC :: Int -> Colist x -> [x]
prefixC 0 s                = []
prefixC n (CL (Left (x,s))) = x:prefixC (n-1) s
prefixC _ _    	           = []

instance Show x => Show (Colist x) where show = show . prefixC 100

xlist :: ColistSig x [x]			         -- final Colist-algebra 
xlist = CLS $ \case x:s -> Left (x,s)			 -- with term carrier
	            _   -> Right ()
				   
unfoldList :: ColistSig x val -> val -> [x]                 
unfoldList alg val = either (\(x,val) -> x:unfoldList alg val) (const []) 
			    $ split alg val
	 
-- Coinductive definitions to Colist x

zipWithC :: (x -> y -> z) -> Colist x -> Colist y -> Colist z
zipWithC f s s' = case splitC s of 
		  Left (x,s) -> case splitC s' of 
		                Left (y,s') -> consC (f x y) $ zipWithC f s s'
		                Right _ -> nilC
		  Right _ -> nilC
		                
zipWithCI f s = zipWithC f (toColist s) . toColist 

conc :: Colist x -> Colist x -> Colist x
conc s s' = case splitC s of 
	         Left (x,s) -> consC x $ conc s s'
	         Right _ -> case splitC s' of Left (x,s') -> consC x $ conc s s'
		                              Right _ -> nilC
		                
concI s = conc (toColist s) . toColist 

-- BINARY TREES

data BinSig a val = BIS {empty_ :: val, bjoin :: a -> val -> val -> val}
								-- signature
data Bintree a = Empty | Bjoin a (Bintree a) (Bintree a)   	-- terms
          
foldBin :: BinSig a val -> Bintree a -> val			-- fold
foldBin alg Empty                = empty_ alg
foldBin alg (Bjoin a left right) = bjoin alg a (foldBin alg left)
                                               (foldBin alg right)        

leaf :: a -> Bintree a
leaf a = Bjoin a Empty Empty

instance Functor Bintree where
         fmap f (Bjoin a left right) = Bjoin (f a) (fmap f left) $ fmap f right
         fmap _ _ = Empty

instance Comonad Bintree where				
         extract (Bjoin a _ _) = a
         f <<= Empty                  = Empty
         f <<= t@(Bjoin _ left right) = Bjoin (f t) (f <<= left) $ f <<= right

baltree :: [a] -> Bintree a
baltree [] = Empty
baltree s  = case splitAt (length s`div`2) s of 
		  (s1,a:s2) -> Bjoin a (baltree s1) $ baltree s2
	          (a:s,_) -> Bjoin a Empty $ baltree s 
	          
baltree1 = baltree [1..22]

     -- > 12(6(3(2(1),5(4)),9(8(7),11(10))),18(15(14(13),17(16)),21(20(19),22)))

foldBinM :: BinSig a val -> Bintree a -> Id val
foldBinM alg Empty               = return $ empty_ alg
foldBinM alg (Bjoin a left right) = liftM2 (bjoin alg a) (foldBinM alg left)
                                                         (foldBinM alg right)
         
btree1 = Bjoin 6 (Bjoin 7 (Bjoin 11 (leaf 55) $ leaf 33) $ Empty) $ leaf 9

sumBA = BIS 0 (\i j k ->i+j+k)	 

btree2 = foldBin sumBA btree1	     
btree3 = (run . foldBinM sumBA) btree1  	      -- > 121	

btree4 = foldBin sumBA <<= btree1 :: Bintree Int 
btree5 = run . foldBinM sumBA <<= btree1 :: Bintree Int 
						      -- > 121(106(99(55,33)),9)
takeTree :: Int -> Bintree a -> Bintree a
takeTree 0 _                    = Empty
takeTree n (Bjoin a left right) = Bjoin a (takeTree (n-1) left) 
					  (takeTree (n-1) right)
takeTree _ Empty                = Empty

addBT :: Num x => Bintree x => Bintree x => Bintree x
addBT Empty Empty = Empty
addBT t Empty     = t
addBT Empty t     = t
addBT (Bjoin x t1 u1) (Bjoin y t2 u2) = Bjoin (x+y) (addBT t1 t2) (addBT u1 u2)

powT :: Num x => x => Bintree x
powT x = Bjoin x t t where t = powT $ x+x

btree6 = takeTree 5 $ powT 1 

         -- > 1(2(4(8(16,16),8(16,16)),4(8(16,16),8(16,16))),
         --     2(4(8(16,16),8(16,16)),4(8(16,16),8(16,16))))

natsT :: Num x => Bintree x
natsT = Bjoin 1 (add $ powT 1) (add $ powT 2) where add = addBT natsT
         
btree7 = takeTree 5 natsT 
         
         -- > 1(2(4(8(16,17),9(18,19)),5(10(20,21),11(22,23))),
         --     3(6(12(24,25),13(26,27)),7(14(28,29),15(30,31))))

mirrorT :: Bintree x => Bintree x
mirrorT Empty         = Empty
mirrorT (Bjoin x t u) = Bjoin x (mirrorT u) (mirrorT t)
         
btree8 = takeTree 5 $ mirrorT natsT

	 -- > 1(3(7(15(31,30),14(29,28)),6(13(27,26),12(25,24))),
	 --     2(5(11(23,22),10(21,20)),4(9(19,18),8(17,16))))

getPath,getPathM :: (Ord x,MonadPlus m) => x -> Bintree x -> m Node

getPath _ Empty         = mzero
getPath x (Bjoin y t u) = if x == y then return []
			  else if x < y 
			       then do node <- getPath x t; return $ 0:node
			       else do node <- getPath x u; return $ 1:node
			       
getPathM _ Empty         = mzero
getPathM x (Bjoin y t u) = msum [do guard $ x == y; return [],
			         do guard $ x < y; node <- getPathM x t
			            return $ 0:node,
			         do node <- getPathM x u; return $ 1:node]

path1 = (getPath 7 baltree1)   :: [Node] 	-- > [[0,1,0,0]]
path2 = (getPath 17 baltree1)  :: Maybe Node	-- > Just [1,0,1]
path3 = (getPathM 23 baltree1) :: Maybe Node	-- > Nothing

-- Bintree parser and unparser

bintree :: Compiler x -> Compiler (Bintree x)
bintree comp = do x <- comp
                  msum [do tchar '('; t <- bintree comp
                           tchar ')'; return $ Bjoin x t Empty,
                        do tchar '('; t <- bintree comp
                           tchar ','; u <- bintree comp
                           tchar ')'; return $ Bjoin x t u,
                        return $ leaf x]

enclose str = ('(':) . str . (')':)

instance Show a => Show (Bintree a) where 
	 showsPrec _ Empty                 = id
	 showsPrec _ (Bjoin a Empty Empty) = shows a 
	 showsPrec _ (Bjoin a t Empty)     = shows a . enclose (shows t)
	 showsPrec _ (Bjoin a Empty t)     = shows a . enclose (shows t)
	 showsPrec _ (Bjoin a t u)         = shows a . enclose
	   				            (shows t . (',':) . shows u)
{-
	 showsPrec _ Empty         = ("Empty"++) 
	 showsPrec _ (Bjoin a t u) = ("Bjoin("++) . shows a . (',':) . shows t
	   					  . (',':) . shows u . (')':)
-}		  	 				     

-- Cobintrees

data CobinSig x val = CBS {splitB :: val -> Either (x,val,val) ()}
					 	      -- signature
newtype Cobintree x = CB {splitBC :: Either (x,Cobintree x,Cobintree x) ()}  
						      -- coterms
cobintreeC :: CobinSig x (Cobintree x)	      	      -- coterm algebra
cobintreeC = CBS splitBC
						     
unfoldCobintree :: CobinSig x val -> val -> Cobintree x
unfoldCobintree alg val = CB $ either (\(x,t,u) -> Left (x,unfold t,unfold u))
			              (const $ Right ()) $ splitB alg val 
			  where unfold = unfoldCobintree alg

bintreeC :: CobinSig x (Bintree x)		      -- final Cobintree-algebra
bintreeC = CBS $ \case Empty -> Right ()	      -- with term carrier
		       Bjoin x t u -> Left (x,t,u)
							   
unfoldBintree :: CobinSig x val -> val -> Bintree x   
unfoldBintree alg val = either (\(x,t,u) -> Bjoin x (unfold t) (unfold u))
			       (const Empty) $ splitB alg val
			where unfold = unfoldBintree alg

emptyC :: Cobintree x
emptyC = CB $ Right ()

bjoinC :: x -> Cobintree x -> Cobintree x -> Cobintree x
bjoinC x t u = CB $ Left (x,t,u)

leafC :: x -> Cobintree x
leafC x = bjoinC x emptyC emptyC

-- Coinductive definitions to Cobintree x

addBC :: Num x => Cobintree x -> Cobintree x -> Cobintree x
addBC t u = case splitBC t of 
		 Left (x,t1,t2) -> case splitBC u of 
		                   Left (y,u1,u2) -> bjoinC (x+y) (addBC t1 u1)
		                   				  (addBC t2 u2)
		                   _ -> t
		 _ -> case splitBC u of 
		                   Right () -> emptyC
		                   _ -> u

powC :: Num x => x -> Cobintree x
powC x = CB $ Left (x,t,t) where t = powC $ x+x

natsC :: Num x => Cobintree x
natsC = CB $ Left (1,add $ powC 1,add $ powC 2) where add = addBC natsC

mirrorC :: Cobintree x => Cobintree x
mirrorC t = case splitBC t of 
                 Left (x,t1,t2) -> bjoinC x (mirrorC t2) (mirrorC t1)
		 _ -> emptyC

btree10 = mirrorC natsC

-- Cobintree parser and unparser

cobintree :: Compiler x -> Compiler (Cobintree x)
cobintree comp = do x <- comp
                    msum [do tchar '('; t <- cobintree comp
                             tchar ')'; return $ bjoinC x t emptyC,
                          do tchar '('; t <- cobintree comp
                             tchar ','; u <- cobintree comp
                             tchar ')'; return $ bjoinC x t u,
                          return $ leafC x]

toBintree :: Int -> Cobintree x -> Bintree x
toBintree 0 t                   = Empty
toBintree n (CB (Left (x,t,u))) = Bjoin x (f t) (f u) where f = toBintree $ n-1
toBintree _ _     		= Empty

toCobintree :: Bintree x -> Cobintree x
toCobintree (Bjoin x t u)         = bjoinC x (toCobintree t) (toCobintree u)
toCobintree Empty	          = emptyC

instance Show x => Show (Cobintree x) where show = show . toBintree 11

btree11 = "32(25(17,30(8(6,99))),50(42(55,66),63(89)))"

doAndShowB :: (Cobintree Int -> Cobintree Int) -> String -> Maybe (Cobintree Int)
doAndShowB f str = do t <- parse (cobintree nat) str; Just $ f t
                     
btree12 = doAndShowB (subtreeC [0,1,0]) btree11	-- > Just 8(6,99)             

doAndShow :: Show x => Compiler x -> [x -> x] -> String -> IO (Maybe ())
doAndShow comp fs str = runMT $ do t <- liftT $ parse comp str
		                   lift $ mkTrees "trees" $ tail $ init 
		                        $ show $ t:map ($t) fs

btree13 = doAndShow (cobintree nat) (map subtreeC [[0,1,0],[0,1,0,1],[1,0]]) 
		    btree11
btree14 = doAndShow (cobintree nat) [subtreeC [0,1,0,0]] $ show baltree1
btree15 = doAndShow (cobintree nat) [] $ show $ baltree [1..100]
btree16 = doAndShow (bintree nat) [] $ show $ baltree [1..100]

-- TREES

data TreeSig a val = TS {var :: a -> val,			-- signature
                         fun :: a -> [val] -> val}

foldTree :: TreeSig a val -> Tree a -> val			-- fold
foldTree alg (V a)    = var alg a
foldTree alg (F a ts) = fun alg a $ map (foldTree alg) ts

label :: Tree a -> Node -> a
label t [] 				= root t
label (F _ ts) (i:node) | i < length ts = label (ts!!i) node
label _ _  				= error "label"

lab1 = label tree1 [4]  -- >  7

-- getSubtree(t)(node) is the subtree of t whose root is node. 

getSubtree :: Node -> Tree a -> Tree a
getSubtree [] t = t
getSubtree (i:node) (F _ ts) | i < length ts = getSubtree node $ ts!!i
getSubtree _ _  			     = error "getSubtree"

depthfirst :: Tree a -> [a]
depthfirst t = root t:concatMap depthfirst (subtrees t)

breadthfirst :: [Tree a] -> [a]
breadthfirst [] = []
breadthfirst ts = map root ts ++ breadthfirst (concatMap subtrees ts)

single a = [a]

-- preordLA(t) lists the nodes of t in preorder.

preordLA :: TreeSig a [a]
preordLA  = TS single $ \a ass -> a:concat ass

-- postordLA(t) lists the nodes of t in postorder.

postordLA :: TreeSig a [a]
postordLA  = TS single $ \a ass -> concat ass++[a]

sumA :: Num a => TreeSig a a
sumA = TS id $ \a as -> a+sum as

arithA :: TreeSig String Int 
arithA = TS {var = \case "x" -> 5; "y" -> -66; "z" -> 13,
             fun  = \case "+" -> sum; "*" -> product}

-- preordT(t) labels each node of t with its position in preordLA(t).

preordT :: Tree a -> Int -> (Tree Int,Int)
preordT (F _ ts) i = (F i us,k) where (us,k)= preordTL ts $ i+1
preordT _ i        = (V i,i+1)

preordTL :: [Tree a] -> Int -> ([Tree Int],Int)
preordTL (t:ts) i = (u:us,k) where (u,j)  = preordT t i; (us,k) = preordTL ts j
preordTL _ i      = ([],i) 

nextPos :: State Int Int
nextPos = State $ \i -> (i,i+1)

preordTM :: Tree a -> State Int (Tree Int)
preordTM (F _ ts) = do i <- nextPos; ts <- preordTLM ts; return $ F i ts
preordTM _        = do i <- nextPos; return $ V i

preordTLM :: [Tree a] -> State Int [Tree Int]
preordTLM (t:ts) = do t <- preordTM t; ts <- preordTLM ts; return $ t:ts
preordTLM _      = return [] 

preordTA :: TreeSig a (State Int (Tree Int))
preordTA = TS {var  = \_    -> do i <- nextPos; return $ V i,
	       fun  = \_ ms -> do i <- nextPos; ts <- sequence ms
	       		 	  return $ F i ts}
		   
tree1,tree2 :: Tree Int
tree1 = F 1 [F 2 [F 2 [V 3,V(-1)],V(-2)],F 4 [V(-3),V 5]]
tree2 = F 11 $ map (\a -> F a [V $ a+1]) [3..11] 

tree3 :: Tree String
tree3 = F "+" [F "*" [V "x",V "y"], V "z"]     

tree4 = foldTree sumA tree2		-- > 146

tree5 = foldTree preordLA tree1		-- > [1,2,2,3,-1,-2,4,-3,5]
					-- depthfirst tree1

tree6 = foldTree postordLA tree1	-- > [3,-1,2,-2,2,-3,5,4,1]

tree7 = foldTree arithA tree3		-- > -317

tree8 = fst $ preordT tree2 0
tree9 = fst $ runS (foldTree preordTA tree2) 0	
	 	   -- > F 0 [F 1 [V 2],F 3 [V 4],F 5 [V 6],F 7 [V 8],F 9 [V 10],
	 	   --        F 11 [V 12],F 13 [V 14],F 15 [V 16],F 17 [V 18]]

tree10 = fst $ preordT tree3 0 	
tree11 = fst $ runS (preordTM tree3) 0 	
tree12 = fst $ runS (foldTree preordTA tree3) 0	-- > F 0 [F 1 [V 2,V 3],V 4]

instance Functor Tree where fmap = mapTree
         
instance Comonad Tree where extract = root
        		    f <<= t@(V _)    = V $ f t
        		    f <<= t@(F _ ts) = F (f t) $ map (f <<=) ts

tree13 = foldTree sumA <<= tree2 	
               -- > F 146 [F 7 [V 4],F 9 [V 5],F 11 [V 6],F 13 [V 7],F 15 [V 8],
               --          F 17 [V 9],F 19 [V 10],F 21 [V 11],F 23 [V 12]]

tree14 = foldTree arithA <<= tree3  -- >  F (-317) [F (-330) [V 5,V (-66)],V 13]

balanced (F _ (t:ts)) = and $ map f ts where f u = height t == height u
balanced _            = True

tree15 = balanced <<= tree3	    -- > F False [F True [V True,V True],V True]

-- Trees with a position

infixr 4 :&

data TreeNode a = (:&) {tree :: Tree a, node :: Node}	     
         
mkTN :: Tree a -> TreeNode a
mkTN t = t:&[]
         
labelTN :: TreeNode a -> a
labelTN (t:&node) = label t node

children :: TreeNode a -> [Tree a]
children (t:&node) = subtrees $ getSubtree node t

rootTN,leafTN,fstchild :: TreeNode a -> Bool
rootTN (_:&node)   = null node
leafTN             = null . children
fstchild (_:&node) = last node == 0

parent,prevchild,lastchild :: TreeNode a -> TreeNode a
parent (t:&node)       = t:&init node
prevchild (t:&node)    = t:&init node++[last node-1]
lastchild tn@(t:&node) = t:&node++[lg-1] where lg = length $ children tn

instance Functor TreeNode where fmap f (t:&node) = mapTree f t:&node

nodeTree :: Tree a -> Node -> Tree Node
nodeTree (V _) node    = V node
nodeTree (F _ ts) node = F node $ zipWith f ts [0..length ts-1]
			 where f t i = nodeTree t $ node++[i]

tree16 = nodeTree tree2 []
            	 -- > F [] [F [0] [V [0,0]],F [1] [V [1,0]],F [2] [V [2,0]],
            	 --         F [3] [V [3,0]],F [4] [V [4,0]],F [5] [V [5,0]],
            	 --         F [6] [V [6,0]],F [7] [V [7,0]],F [8] [V [8,0]]]
                
instance Comonad TreeNode where	extract (t:&node) = label t node
                 		f <<= t:&node = mapTree (f . (t:&)) nt:&node 
         			                where nt = nodeTree t []
         
prefixSumTN :: TreeNode Int -> Int
prefixSumTN tn | rootTN tn = root $ tree tn
               | True      = prefixSumTN (parent tn)+labelTN tn

tree17 = tree $ prefixSumTN <<= mkTN tree2
         -- > F 11 [F 14 [V 18],F 15 [V 20],F 16 [V 22],F 17 [V 24],F 18 [V 26],
         --         F 19 [V 28],F 20 [V 30],F 21 [V 32],F 22 [V 34]]
	  
depthTN,breadthTN :: TreeNode a -> Int 			
depthTN (t:&node)   = fromJust $ lookup node 
                               $ zip (depthfirst $ nodeTree t []) [0..]
breadthTN (t:&node) = fromJust $ lookup node 
			       $ zip (breadthfirst [nodeTree t []]) [0..]
			       
-- The following version of depthTN has been inspired by the numin/numout 
-- example of T. Uustalu, V. Vene, Comonadic functional attribute evaluation, 
-- in: M. van Eekelen, ed., Trends in Functional Programming 6, Intellect (2007) 
-- 145-162. 

depthUV :: TreeNode a -> Int 
depthUV tn | rootTN tn   = 0
           | fstchild tn = depthUV (parent tn)+1
           | True        = aux (prevchild tn)+1
           		   where aux tn | leafTN tn = depthUV tn 
	                                | True      = aux $ lastchild tn

tree18 = tree $ depthUV <<= mkTN tree2
tree19 = tree $ depthTN <<= mkTN tree2
	 	   -- > F 0 [F 1 [V 2],F 3 [V 4],F 5 [V 6],F 7 [V 8],F 9 [V 10],
	 	   --        F 11 [V 12],F 13 [V 14],F 15 [V 16],F 17 [V 18]]

tree20 = tree $ depthUV <<= mkTN tree3	
tree21 = tree $ depthTN <<= mkTN tree3	-- > F 0 [F 1 [V 2,V 3],V 4]

tree22 = tree $ breadthTN <<= mkTN tree2

tree23 = tree $ breadthTN <<= mkTN tree3

-- AUTOMATA ****

-- Deterministic automata

data DAut x y state = DAut {delta :: state -> x -> state,   -- signature
			    beta :: state -> y}

data DAutC x y = New {deltaC :: x -> DAutC x y,	            -- DAut-coterms
		      betaC :: y}     			    -- coDAut-terms

dAutC :: DAut x y (DAutC x y)			            -- (co)term algebra
dAutC = DAut deltaC betaC				    

esum,osum :: DAutC Int Bool				    -- two (co)terms
esum = New (\x -> if even x then esum else osum) True
osum = New (\x -> if even x then osum else esum) False

unfoldDAut :: DAut x y state -> state -> DAutC x y	    -- unfold
unfoldDAut alg s = New (unfoldDAut alg . delta alg s) $ beta alg s

behFun :: DAut x y ([x] -> y)			            -- another final 
behFun = DAut (\f x -> f . (x:)) ($ [])                     -- DAut-algebra

unfoldBeh :: DAut x y state -> state -> [x] -> y            -- corresponding
unfoldBeh alg state = beta alg . foldl (delta alg) state    -- unfold

eo1 = unfoldBeh dAutC esum			    	    -- > even . sum 
eo2 = unfoldBeh dAutC osum			    	    -- > odd . sum 

data EO = Esum | Osum deriving Eq

eo :: DAut Int Bool EO			                    -- automaton 
eo = DAut (\case Esum -> f . even			    -- with two states
                 Osum -> f . odd) (== Esum) 
     where f b = if b then Esum else Osum
     
eo3 = unfoldBeh eo Esum 				    -- > even . sum 
eo4 = unfoldBeh eo Osum 				    -- > odd . sum 

-- Monadic automata

-- class MonadEq m where unit :: a -> m a
		      -- bind :: (Eq a,Eq b) => m a -> (a -> m b) -> m b		     

data MAut m x y s = MAut {ini :: s, deltaM :: s -> x -> m s,  
			  betaM :: s -> y}		      -- signature

reach :: Monad m => MAut m x y s -> [x] -> s -> m s
reach _ [] s      = return s
reach alg (x:w) s = deltaM alg s x >>= reach alg w

obs :: (Functor m,Monad m) => MAut m x y s -> s -> [x] -> m y
obs alg s w = fmap (betaM alg) $ reach alg w s

beh :: (Functor m,Monad m) => MAut m x y s -> [x] -> m y
beh alg = obs alg $ ini alg

-- Nondeterministic automata

type NDAut = MAut [ ]

{- data Set a = S {unS :: [a]} deriving Eq		   -- powerset monad

instance MonadEq Set where unit a = S [a]
			   S s `bind` f = S $ mkSet $ concatMap (unS . f) s
				     
mkSet :: Eq a => [a] -> [a]
mkSet = foldl addS []

addS :: Eq a => [a] -> a -> [a]
addS s@(a:t) b = if a == b then s else a:addS t b
addS _ a       = [a]
-}

-- Probabilistic automata

data Dist a = D {runD :: [(a,Float)]}			   -- distribution monad

instance Functor Dist where
         fmap f (D d) = D [(f a,p) | (a,p) <- d]
 	 
instance Applicative Dist where pure  = return
         			(<*>) = liftM2 id
         
instance Monad Dist where 
         return a = D [(a,1)]
	 D d >>= f = D [(b,p*q) | (a,p) <- d, (b,q) <- runD $ f a]
	 fail _ = D []

plusD :: Eq a => Dist a -> Dist a -> Dist a
D d `plusD` D d' = D $ foldl addD d d'

addD :: Eq a => [(a,Float)] -> (a,Float) -> [(a,Float)]
addD (ap@(a,p):s) bq@(b,q) = if a == b then (a,p+q):s else ap:addD s bq
addD _ ap                  = [ap]

-- Examples from Erwig, Kollmansberger, Probabilistic Functional Programming
-- in Haskell

float :: RealFloat a => Int -> a
float = fromInteger . toInteger

uniform :: [a] -> Dist a
uniform s = D [(a,1/total) | a <- s] where total = float $ length s

dice = [1..6]

two6 :: MAut Dist () Bool [Int]
two6 = MAut [] (\s -> const $ uniform [s++[n] | n <- dice])
	       (\s -> length (filter (== 6) s) >= 2)

data Marble = R | G | B deriving Eq

rgb :: MAut Dist () Bool ([Marble],[Marble])
rgb = MAut ([R,R,G,G,B],[]) f ((== [R,G,B]) . snd)
      where f (jar,drawn) _ = uniform [(delete m jar,drawn++[m]) | m <- jar]

doors = [1..3]

data HallState = HS {prize,chosen,opened :: Int} deriving Eq

data Action = Hide | Choose | Open | Switch Bool

game :: MAut Dist Action Bool HallState
game = MAut (HS {prize = undefined, chosen = undefined, opened = undefined})
            (\s -> uniform .
	                \case 
	                  Hide -> [s{prize=d} | d <- doors]
		          Choose -> [s{chosen=d} | d <- doors]
		          Open -> [s{opened=d} | d <- doors\\[prize s,chosen s]]
		          Switch True 
			    -> [s{chosen=d} | d <- doors\\[opened s,chosen s]]
		          _ -> [s])
            (\s -> chosen s == prize s)

dexa :: Int -> [(Bool,Float)]
dexa 1 = runD $ beh two6 $ replicate 4 ()
         -- > [(False,0.868048),(True,0.13194445)]
dexa 2 = runD $ beh rgb $ replicate 3 ()
         -- > [(False,0.93333346),(True,6.666667e-2)]
dexa 3 = runD $ beh game [Hide,Choose,Open,Switch True] 
         -- > [(False,0.33333334),(True,0.6666667)]
dexa _ = runD $ beh game [Hide,Choose,Open,Switch False]
         -- > [(True,0.33333334),(False,0.6666667)]
         
-- Further automata are implemented in Compiler.hs.


-- FUNCTOR (CO)ALGEBRAS

data HList a s = HNil | HCons (a,HList a s)		    -- list functor

data TList a = InList (HList a (TList a))		    -- list terms

listH :: HList a [a] -> [a]				    -- a list algebra
listH HNil          = []
listH (HCons (a,s)) = a:listH s

data HStream a s = HStream {hd_ :: a,tl_ :: HStream a s}    -- stream functor

data StreamF a = InStream {finAlg :: HStream a (StreamF a)} -- stream coterms

stream :: [a] -> HStream a [a]				    -- stream coalgebra
stream (a:s) = HStream {hd_ = a,tl_ = stream s}
