module Movie where

-- 9.7.2022

import System.Process
import System.Directory
import Data.Maybe (fromJust)

infixl 6 &

search :: (a -> Bool) -> [a] -> Maybe Int
search f = g 0 where g i (a:s) = if f a then Just i else g (i+1) s
                     g _ _     = Nothing
		      
getInd :: Eq a => [a] -> a -> Int
getInd s a = fromJust $ search (== a) s

(&) :: String -> String -> String
x & val = x ++ "=\"" ++ val ++ "\""
 
mkFile dir i | i < 10  = dir ++ "/00" ++ file
	     | i < 100 = dir ++ "/0"  ++ file
	     | True    = dir ++ '/'   :  file where file = show i ++ ".svg"

html :: Bool -> String -> [String] -> IO ()
html frame dir files@(first:rest) = writeFile (dir ++ ".html") $
    "<html>\n<head>\n<title>" ++ dir ++ 
    "</title>\n<script type"&"text/javascript" ++ 
    ">\nvar images = new Array(" ++ '\"':first ++ '\"':concatMap f rest ++ 
    ")\n</script>\n<script type"&"text/javascript" ++ " src"&"Painter.js" ++ 
    ">\n</script>\n</head>\n<body style"&"background-color: rgb(221,221,255)" ++
    ">\n<input type"&"button" ++ " value"&"|<<" ++ " onclick"&"backStop(this)"++ 
    ">\n<input type"&"button" ++ " value"&"<" ++ " onclick"&"back()" ++ "> " ++ 
    show n ++ " file" ++ (if n == 1 then "" else "s") ++
    "\n<input type"&"button" ++ " value"&">" ++ " onclick"&"forth()" ++ 
    ">\n<input type"&"button" ++ " value"&">>|" ++" onclick"&"forthStop(this)"++ 
    ">\n<input type"&"button" ++ " value"&"loop" ++ " onclick"&"loop(this)" ++ 
    ">\n<br><input id"&"fileSlider" ++ " type"&"range" ++ " min"&"0" ++ 
    " max"&show (n-1) ++ " value"&"0" ++ " onchange"&"showFile(this.value)" ++
    " style"&"width:800px" ++ ">\n<span id"&"file" ++ '>':first ++ 
    "</span>\n<br>\n<input type"&"range" ++ " min"&"0" ++ " max"&"3000" ++ 
    " value"&"30" ++ " onchange"&"showTime(this.value)" ++ 
    " style"&"width:800px" ++ ">\n<span id"&"time" ++ 
    ">30</span> millisecs\n<br><br><br>\n<" ++ open ++ " id"&"img" ++ 
    " src"&first ++ " width"&width ++ " height"&height ++ close ++ 
    ">\n</body>\n</html>"
  where n = length files
        f file = ",\"" ++ file ++ "\""
	(open,width,height,close) = 
	 if frame then ("iframe","1800","900"," frameborder"&"0"++">\n</iframe")
		  else ("img","600","600",">\n</img")
html _ _ _ = return ()

embed :: String -> (String,(Int,Int)) -> String
embed back (code,p) | null back = svgHead p ++ code ++ "</svg>"
	            | True = svgHead p ++ "\n<image xlink:href=\"" ++ back ++ 
	                     '\"':svgFrame p ++ '/':'>':code ++ "</svg>"
                          -- '\n':code ++ "<image xlink:href=\"" ++ back ++ 
                          -- '\"':svgFrame p ++ "/></svg>" -- back is foreground

svgHead (x,y) = "<svg xmlns"&"http://www.w3.org/2000/svg" ++
	        "\nxmlns:xlink"&"http://www.w3.org/1999/xlink" ++
		"\nversion"&"1.1" ++ " baseProfile"&"full" ++
		"\nwidth"&show (x+20) ++ " height"&show (y+20) ++ ">"

svgFrame (x,y) = " x"&"5"++" y"&"5"++" width"&show (x+10)++" height"&show (y+10)

border p = "\n<rect" ++ svgFrame p ++ " fill"&"none" ++ " stroke"&
           "rgb(100,100,100)" ++ " stroke-width"&"1.5" ++ "/>"

isPic :: String -> Bool
isPic file = lg > 5 && drop (lg-5) file == ".jpeg" || lg > 4 && 
	     drop (lg-4) file `elem` words ".eps .gif .jpg .JPG .pdf .png .svg"
	     where lg = length file
                  
pix file = "Pix/" ++ file

moviePix :: Bool -> String -> IO ()
moviePix frame dir = do files <- getDirectoryContents $ pix dir
                        html frame (pix dir) $ map f $ filter isPic files
                     where f file = dir ++ '/':file

movies :: String -> [(String,Int)] -> IO ()
movies dir ps = do filess <- mapM getDirectoryContents $ map pix dirs
                   let picFiles dir = filter isPic $ filess!!getInd dirs dir
                       f (dir,i) = dir ++ '/':picFiles dir!!i
                   html False (pix dir) $ map f ps
                where dirs = map fst ps

movieSelect dir = movies dir . map (\i -> (dir,i))
 
mkPdf file = system $ "pstopdf " ++ file

mkPng :: String -> String -> IO ()
mkPng dir file = do rawSystem "convert" [file',"-trim",new]; removeFile file'
		 where file' = dir ++ '/':file
		       new = take (length file'-4) file' ++ ".png"

-- mkPng works for eps files. For converting svg files to png use
-- https://cloudconvert.org/svg-to-png.				

pngmovie :: String -> IO ()
pngmovie dir = do files <- getDirectoryContents dir
	          mapM_ (mkPng dir) $ filter isPic files
                  files <- getDirectoryContents dir
	          html True dir [dir++'/':file | file <- files, isPic file]
	           
getFiles :: String -> IO ()
getFiles dir = do files <- getDirectoryContents dir; mapM_ putStrLn files

